# -*- coding: utf-8 -*-

import time
import re


def get_version():
    patterns = [
            r'^set\(UFO_VERSION_MAJOR "(\d*)"\)',
            r'^set\(UFO_VERSION_MINOR "(\d*)"\)',
            r'^set\(UFO_VERSION_PATCH "(\d*)"\)'
            ]
    version = ["0", "0", "0"]

    with open('../../CMakeLists.txt', 'r') as f:
        lines = f.readlines()
        major_pattern = r'^set\(UFO_VERSION_MAJOR "(\d*)"\)'

        for line in lines:
            for i, pattern in enumerate(patterns):
                m = re.match(pattern, line)

                if m:
                    version[i] = m.group(1)

    return '.'.join(version)


project = u'UFO'
copyright = u'%s, UFO Development Team a Collaboration of KIT, SCI and TPU' % time.strftime('%Y')
version = get_version()
release = version

extensions = [
    'sphinx.ext.imgmath',
]

templates_path = ['_templates']

master_doc = 'index'
source_suffix = '.rst'
source_encoding = 'utf-8'

primary_domain = 'c'

add_function_parentheses = True
pygments_style = 'sphinx'

try:
    import sphinx_rtd_theme

    html_theme = 'sphinx_rtd_theme'
    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]
except ImportError:
    pass

html_logo = "_static/ufo-logo.png"
html_static_path = ['_static']
html_use_smartypants = True

htmlhelp_basename = 'UFOdoc'

latex_documents = [
  ('index', 'UFO.tex', u'UFO Documentation',
   u'Matthias Vogelgesang', 'manual'),
]



from setuptools import setup

setup(
    name='tangofu',
    version='0.0.1',
    author='Matthias Vogelgesang',
    author_email='matthias.vogelgesang@kit.edu',
    url='http://ufo.kit.edu',
    license='(?)',
    description='TANGO server for UFO',
    long_description='TANGO server for UFO',
    scripts=['Ufo'],
    install_requires=['PyTango']
)



import tifffile
from gi.repository import GLib, Ufo
from nose.tools import raises
from common import tempdir, disable


@raises(GLib.GError)
def test_wrong_connection():
    from ufo import Read, FlatFieldCorrect, Write

    darks = Read()
    flats = Read()
    radios = Read()
    write = Write()
    ffc = FlatFieldCorrect()

    g = Ufo.TaskGraph()
    g.connect_nodes_full(radios.task, ffc.task, 0)
    g.connect_nodes_full(darks.task, ffc.task, 1)
    g.connect_nodes_full(flats.task, ffc.task, 0)
    g.connect_nodes(ffc.task, write.task)

    sched = Ufo.Scheduler()
    sched.run(g)


def test_task_count():
    from ufo import DummyData, Write, Average

    with tempdir() as d:
        generate = DummyData(number=5, width=512, height=512)
        write = Write(filename=d.path('foo-%i.tif'))
        average = Average()

        write(average(generate())).run().join()

        assert(generate.task.props.num_processed == 0)
        assert(average.task.props.num_processed == 5)
        assert(write.task.props.num_processed == 1)


@disable
def test_broadcast():
    from ufo import Generate, Writer
    import glob

    with tempdir() as d:
        generate = Generate(number=5, width=512, height=512)
        write1 = Writer(filename=d.path('foo-%i.tif'))
        write2 = Writer(filename=d.path('bar-%i.tif'))

        g = Ufo.TaskGraph()
        g.connect_nodes(generate.task, write1.task)
        g.connect_nodes(generate.task, write2.task)

        sched = Ufo.Scheduler()
        sched.run(g)

        foos = glob.glob(d.path('foo-*'))
        bars = glob.glob(d.path('bar-*'))
        assert(len(foos) == 5)
        assert(len(bars) == 5)


def test_resource_info():
    resources = Ufo.Resources()
    nodes = resources.get_gpu_nodes()
    assert(nodes)
    node = nodes[0]
    assert(node.get_info(Ufo.GpuNodeInfo.LOCAL_MEM_SIZE) > 0)
    assert(node.get_info(Ufo.GpuNodeInfo.MAX_MEM_ALLOC_SIZE) > 0)
    assert(node.get_info(Ufo.GpuNodeInfo.GLOBAL_MEM_SIZE) > node.get_info(Ufo.GpuNodeInfo.LOCAL_MEM_SIZE))



import ufo.numpy
import numpy as np
import tifffile
import contextlib
from common import disable, tempdir


a, b = 1.5, 2.5
ones = np.ones((512, 512))
zeros = np.zeros((512, 512))
small = np.ones((256, 256))
random = np.random.random((512, 512))


def have_camera_plugin():
    from gi.repository import Ufo

    return 'camera' in Ufo.PluginManager().get_all_task_names()


@contextlib.contextmanager
def single_tiff_setup(n_images, fmt='foo-{:05}.tif'):
    with tempdir() as d:
        data = np.ones((512, 512), dtype=np.float32)

        for i in range(n_images):
            tifffile.imsave(d.path(fmt.format(i)), data)

        yield d


def test_read_single_tiffs():
    from ufo import Read, Null

    with single_tiff_setup(32) as d:
        read = Read(path=d.root)
        null = Null()

        null(read()).run().join()
        assert(null.task.props.num_processed == 32)


def test_read_single_tiffs_stepped():
    from ufo import Read, Null

    with single_tiff_setup(32) as d:
        read = Read(path=d.root, step=2)
        null = Null()
        null(read()).run().join()
        assert(null.task.props.num_processed == 32 / 2)


def test_read_single_tiffs_start_modified():
    from ufo import Read, Null

    with single_tiff_setup(32) as d:
        read = Read(path=d.root, start=15)
        null = Null()

        null(read()).run().join()
        assert(null.task.props.num_processed == 32 - 15)


@disable
def test_read_multi_tiffs():
    from ufo import Read, Null

    with tempdir() as d:
        n_images = 32
        data = np.ones((512, 512, n_images))
        tifffile.imsave(d.path('foo.tif'), data)

        read = Read(path=d.path('foo.tif'))
        null = Null()

        null(read()).run().join()
        assert(null.task.props.num_processed == n_images)


def test_average():
    from ufo import Average
    average = Average()

    for x in average([a * ones, b * ones]):
        expected = (a + b) / 2
        assert(np.all(x == expected))


def test_buffer():
    from ufo import DummyData, Buffer

    data = DummyData(number=10, width=512, height=256)
    buffr = Buffer()
    result = list(buffr(data()))
    assert(len(result) == 10)

    for r in result:
        assert(r.shape[0] == 256)
        assert(r.shape[1] == 512)


def test_rescale():
    from ufo import Rescale
    rescale = Rescale(factor=0.5)
    result = list(rescale([a * ones, b * small]))
    assert(np.mean(result[0]) == a)
    assert(np.mean(result[1]) == b)


@disable
def test_roi():
    from ufo import CutRoi

    x, y = 10, 20
    w, h = 256, 128
    roi = CutRoi(x=x, y=y, width=w, height=h)
    result = list(roi([random, random]))
    ref = random[y:y+h, x:x+w]
    assert(ref.shape[0] == h)
    assert(ref.shape[1] == w)
    assert(np.all(ref == result[0]))


def test_stack():
    from ufo import Stack
    stack = Stack(number=2)

    for x in stack([a * ones, b * ones]):
        assert(x.shape[0] == 2)
        assert(np.all(x[0,:,:] == a))
        assert(np.all(x[1,:,:] == b))


def test_flatten():
    from ufo import FlattenInplace

    summing = FlattenInplace(mode="sum")
    result = list(summing([a * ones, b * ones]).items())
    assert(np.all(result[0] == a + b))


def test_fft_1d():
    from ufo import Fft, Ifft

    fft = Fft(dimensions=1)
    ifft = Ifft(dimensions=1)
    orig_a = a * ones
    orig_b = b * random
    result = list(ifft(fft([orig_a, orig_b])))

    assert(np.sum(orig_a - result[0]) < 0.001)
    assert(np.sum(orig_b - result[1]) < 0.01)


def test_fft_2d():
    from ufo import Fft, Ifft

    fft = Fft(dimensions=2)
    ifft = Ifft(dimensions=2)
    orig_a = a * ones
    orig_b = b * random
    result = list(ifft(fft([orig_a, orig_b])))

    assert(np.sum(orig_a - result[0]) < 0.001)
    assert(np.sum(orig_b - result[1]) < 0.1)


def test_flatfield_correction():
    from ufo import FlatFieldCorrect

    darks = np.ones((512, 512)) * 1.5
    flats = np.ones((512, 512)) * 11.5
    projs = np.ones((512, 512)) * 100.0
    ffc = FlatFieldCorrect()

    expected = (projs - darks) / (flats - darks)
    result = list(ffc([projs, projs], [darks, darks], [flats, flats]))[0]
    assert(np.sum(np.abs(expected - result)) < 1)

    expected = - np.log((projs - darks) / (flats - darks))
    ffc = FlatFieldCorrect(absorption_correct=True)
    result = list(ffc([projs, projs], [darks, darks], [flats, flats]))[0]
    assert(np.sum(np.abs(expected - result)) < 1)


def test_measure():
    from ufo import Measure

    measures = []

    def measure_callback(m, a):
        measures.append(ufo.numpy.asarray(a))

    measure = Measure(metric='mean', axis=-1)
    measure.connect('result', measure_callback)
    measure([a * ones, b * ones]).run().join()
    assert(len(measures) == 2)


def test_dummy_data():
    from ufo import DummyData

    data = DummyData(number=10, width=256, height=128)
    result = list(data())
    assert(len(result) == 10)
    assert(all(r.shape[0] == 128 and r.shape[1] == 256 for r in result))


def test_metaballs():
    from ufo import Metaballs

    metaballs = Metaballs(number=5, number_balls=5, width=512, height=256)
    result = list(metaballs())
    assert(len(result) == 5)
    assert(all(r.shape[0] == 256 and r.shape[1] == 512 for r in result))


def test_transpose():
    from ufo import Transpose

    transpose = Transpose()
    ones = np.ones((256, 512))
    zeros = np.zeros((256, 128))
    result = list(transpose([ones, zeros]))
    assert(np.all(result[0] == ones.transpose()))
    assert(np.all(result[1] == zeros.transpose()))


def test_uca():
    if have_camera_plugin():
        from ufo import Camera

        camera = Camera(name='mock', number=2)
        result = list(camera())
        assert(len(result) == 2)


def test_uca_direct():
    try:
        from gi.repository import Ufo, Uca

        if have_camera_plugin():
            from ufo import Camera

            uca_pm = Uca.PluginManager()
            mock = uca_pm.get_camerav('mock', [])
            camera = Camera(camera=mock, count=3)

            result = list(camera())
            assert(len(result) == 3)
    except ImportError:
        pass


def test_memory_in():
    with tempdir() as d:
        from ufo import MemoryIn, Write

        ref = random.astype(np.float32)
        read = MemoryIn(pointer=ref.__array_interface__['data'][0], number=1,
                        width=ref.shape[1], height=ref.shape[0])
        write = Write(filename=d.path('foo.tif'))

        write(read()).run().join()
        result = tifffile.imread(d.path('foo.tif'))
        assert(np.all(ref == result))


def test_memory_out():
    with tempdir() as d:
        from ufo import MemoryOut, Read

        ref = random.astype(np.float32)
        out = np.zeros_like(ref).astype(np.float32)
        tifffile.imsave(d.path('foo.tif'), ref)

        read = Read(path=d.path('foo.tif'))
        write = MemoryOut(pointer=out.__array_interface__['data'][0], max_size=ref.nbytes)

        write(read()).run().join()
        assert(np.all(out == ref))



import numpy as np
import tifffile
from common import tempdir, disable
from gi.repository import Ufo


def test_filters_issue_46_varying_fft_size():
    from ufo import Fft, Ifft

    data = [np.random.random((256, 256)),
            np.random.random((512, 512)),
            np.random.random((1024, 1024))]

    fft = Fft(dimensions=1)
    ifft = Ifft(dimensions=1)
    result = list(ifft(fft(data)))

    for orig, processed in zip(data, result):
        assert(np.sum(orig - processed) < 1.0)


def test_core_issue_61_16_bit_tiffs():
    from ufo import Read, Write

    orig = np.random.randint(0, 65535, (512, 512)).astype(np.uint16)

    with tempdir() as d:
        tifffile.imsave(d.path('temp.tif'), orig)

        read = Read(path=d.path('temp.tif'))
        write = Write(filename=d.path('temp-%i.tif'))

        write(read()).run().join()

        produced = tifffile.imread(d.path('temp-0.tif'))
        assert(np.sum(orig - produced) == 0.0)


@disable
def test_core_issue_64_fixed_expansion():
    g = Ufo.TaskGraph()
    pm = Ufo.PluginManager()
    sched = Ufo.FixedScheduler()
    arch = Ufo.ArchGraph()
    gpus = arch.get_gpu_nodes()
    sched.set_gpu_nodes(arch, gpus)

    generate = pm.get_task('generate')
    null = pm.get_task('null')

    generate.set_properties(number=5, width=512, height=512)

    for gpu in gpus:
        median = pm.get_task('median-filter')
        median.set_proc_node(gpu)
        g.connect_nodes(generate, median)
        g.connect_nodes(median, null)

    sched.run(g)


@disable
def test_core_issue_66_remote_list():
    daemon = Ufo.Daemon.new('tcp://127.0.0.1:5554')
    daemon.start()
    arch = Ufo.ArchGraph(remotes=['tcp://127.0.0.1:5554'])
    daemon.stop()


def test_filters_issue_79():
    from ufo import Stack, Slice

    data = [np.ones((512, 512)) for i in range(10)]

    stack = Stack(number=2)
    unstack = Slice()

    result = list(unstack(stack(data)))
    assert len(result) == len(data)



import os
import sys
import subprocess
import pkgconfig
import numpy
from setuptools import setup, Extension


VERSION = '1.1.0'


if not pkgconfig.installed('pygobject-3.0', '>=3.2.2') and \
   not pkgconfig.installed('pygobject-2.0', '>=2.28'):
    sys.exit("You must install pygobject-2.0 or pygobject-3.0")

if not pkgconfig.installed('ufo', '>=0.4.0'):
    sys.exit("You must install ufo>=0.4.0")

def listify(d):
    return {k: list(v) for k, v in d.items()}

build_flags = listify(pkgconfig.parse('pygobject-3.0 ufo'))

build_flags['include_dirs'].append(numpy.get_include())
build_flags['extra_compile_args'] = ['-std=c99']

setup(
    name='ufo',
    version=VERSION,
    author='Matthias Vogelgesang',
    author_email='matthias.vogelgesang@kit.edu',
    url='http://ufo.kit.edu',
    license='GPL v3 (?)',
    description='ufo extension module',
    long_description='ufo extension module',
    packages=['ufo'],
    ext_modules=[Extension('_ufo', ['src/ufo.c'], **build_flags)],
    zip_safe=False,
)





import sys
import re
import time
import threading
import queue
import numpy as np
import gi
from .numpy import asarray, empty_like
gi.require_version('Ufo', '0.0')
from gi.repository import GObject, Ufo


Ufo.Scheduler()


class Environment(object):

    def __init__(self, pm):
        self.pm = pm
        self.reset()
        self.done = False
        self.started = False

    def reset(self):
        self.graph = Ufo.TaskGraph()


class Task(object):

    def __init__(self, env, name, cargs, task=None):
        self.env = env
        self.name = name
        self.cargs = cargs
        self.task = task or self.env.pm.get_task(self.name)

    def connect(self, *args, **kwargs):
        self.task.connect(*args, **kwargs)

    def __call__(self, *args):
        self.task.set_properties(**self.cargs)

        if len(args) == 0:
            return self

        if any(isinstance(a, Task) for a in args):
            for i, source in enumerate(args):
                self.env.graph.connect_nodes_full(source.task, self.task, i)
        else:
            def input_data(*iargs):
                tasks = []
                buffers = []

                for i in range(len(iargs)):
                    task = Ufo.InputTask()
                    tasks.append(task)
                    buffers.append(None)
                    self.env.graph.connect_nodes_full(task, self.task, i)

                while not self.env.started:
                    time.sleep(0.01)

                for args in zip(*iargs):
                    for i, (task, data) in enumerate(zip(tasks, args)):
                        if buffers[i] is None:
                            buffers[i] = empty_like(data)
                        else:
                            buffers[i] = task.get_input_buffer()

                        buffers[i].set_host_array(data.__array_interface__['data'][0], False)
                        task.release_input_buffer(buffers[i])

                for task in tasks:
                    task.stop()

            thread = threading.Thread(target=input_data, args=args)
            thread.start()

        return self

    def __iter__(self):
        return self.items()

    def run(self, **props):
        """
        Run the tasks.

        Args:
            props (dict): parameter dict used to create the scheduler.
        """

        sched = Ufo.Scheduler()
        sched.set_properties(**props)

        def run_scheduler():
            self.env.done = False
            sched.run(self.env.graph)
            self.env.done = True
            self.env.reset()

        self.thread = threading.Thread(target=run_scheduler)
        self.thread.start()
        self.env.started = True
        return self

    def join(self):
        self.thread.join()

    def items(self):
        results = queue.Queue()

        def processed(task):
            buf = output_task.get_output_buffer()
            results.put(asarray(buf).copy())
            output_task.release_output_buffer(buf)

        output_task = Ufo.OutputTask()
        output_task.connect('processed', processed)
        self.env.graph.connect_nodes(self.task, output_task)
        self.run()

        while not self.env.done or not results.empty():
            try:
                yield results.get(True, 0.01)
            except queue.Empty:
                pass

        self.join()

    @property
    def graph(self):
        return self.env.graph


class TaskLoader(object):

    def __init__(self, env, name):
        self.env = env
        self.name = name

    def __call__(self, **kwargs):
        # This should probably be done with using metaclasses, but it looks
        # simple and straightforward enough
        return Task(self.env, self.name, kwargs)


def get_path(fqn):
    return fqn.split('.')[1]


class Importer(object):
    def __init__(self):
        self.env = Environment(Ufo.PluginManager())
        self.task_names = self.env.pm.get_all_task_names()

    def find_module(self, fqn, path=None):
        if fqn.startswith('ufo.'):
            if get_path(fqn).lower() in self.task_names:
                return self

        return None

    def load_module(self, fqn):
        # Splits according to http://stackoverflow.com/a/1176023/997768
        s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', get_path(fqn))
        task_name = re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()
        return TaskLoader(self.env, task_name)


sys.meta_path += [Importer()]



from fabric.api import run, local
from fabric.context_managers import hide


def start_ufod(cenv=None, limit=None):
    exports = ""
    n_gpus = 0

    with hide('stdout'):
        smi_output = run('nvidia-smi -L')
        n_gpus = len(smi_output.split('\n'))

    n_gpus = min(n_gpus, int(limit)) if limit else n_gpus

    if cenv:
        cenv_path = "$HOME/.local/share/cenv/{}".format(cenv)
        exports = "PATH=$PATH:{cenv}/bin LD_LIBRARY_PATH={cenv}/lib".format(cenv=cenv_path)

    addresses = ['tcp://141.52.65.18:{}'.format(6666 + i) for i in range(n_gpus)]

    for i in range(n_gpus):
        address = 'tcp://141.52.65.18:{}'.format(6666 + i)
        run("{} bash -c '((nohup ufod -l {} > /dev/null 2> /dev/null) &)'".format(exports, address), pty=False)

    return addresses


def start(cmd='ufo-launch', args=None, cenv=None, limit=None):
    addresses = start_ufod(cenv=cenv, limit=limit)
    cmd_line = '{} {}'.format(cmd, ' '.join('-a {}'.format(a) for a in addresses))

    if args:
        cmd_line = '{} {}'.format(cmd_line, args)

    local(cmd_line)




import math
import numpy as np


class Parameters(object):

    def __init__(self, data):
        self.num_sinograms = data.shape[0]
        self.num_projections = data.shape[1]
        self.width = data.shape[2]
        self.data = data

    def __repr__(self):
        return 'Parameters(n_sinos={}, n_proj={}, width={})'.format(self.num_sinograms, self.num_projections, self.width)

    @property
    def sinograms(self):
        return (self.data[i,:,:] for i in range(self.num_sinograms))


def _fbp(data, center, volume, theta, kwargs):
    from ufo import Fft, Ifft, Filter, Backproject

    p = Parameters(data)
    theta = theta[1] - theta[0]
    center = np.mean(center)

    bp = Backproject(axis_pos=center, angle_step=theta, angle_offset=np.pi)
    fft = Fft()
    ifft = Ifft(crop_width=p.width)
    flt = Filter()

    for i, slce in enumerate(bp(ifft(flt(fft(p.sinograms))))):
        volume[i,:,:] = slce


def fbp(*args, **kwargs):
    """
    Reconstruct object using UFO's filtered backprojection algorithm

    Note
    ----
    You have to set ``ncore`` to 1 because UFO provides multi-threading on its
    own.
    """

    # Erm ... okay?
    result = [_fbp]
    result.extend(args)
    result.append(kwargs)
    return result


def _dfi(data, center, volume, theta, kwargs):
    from ufo import Crop, Zeropad, Fft, Ifft, DfiSinc, Ifft, SwapQuadrants

    oversampling = kwargs['options'].get('oversampling', 1)
    p = Parameters(data)
    theta = theta[1] - theta[0]
    center = np.mean(center)
    padded_size = pow(2, int(math.ceil(math.log(p.width, 2))))
    frm = padded_size / 2 - p.width / 2
    to = padded_size / 2 + p.width / 2

    pad = Zeropad(oversampling=1, center_of_rotation=center)
    fft = Fft(dimensions=1, auto_zeropadding=False)
    dfi = DfiSinc(angle_step=theta)
    ifft = Ifft(dimensions=2)
    swap_forward = SwapQuadrants()
    swap_backward = SwapQuadrants()

    for i, slce in enumerate(swap_backward(ifft(swap_forward(dfi(fft(pad(p.sinograms))))))):
        volume[i,:,:] = slce[frm:to, frm:to]


def dfi(*args, **kwargs):
    """
    Reconstruct object using UFO's direct Fourier inversion algorithm

    Extra options
    -------------
    oversampling : int, optional
       Oversampling factor.

    Note
    ----
    You have to set ``ncore`` to 1 because UFO provides multi-threading on its
    own.
    """
    result = [_dfi]
    result.extend(args)
    result.append(kwargs)
    return result


def _ir(data, center, volume, theta, kwargs):
    from ufo import Null, Task
    p = Parameters(data)
    theta = theta[1] - theta[0]
    center = np.mean(center)

    null = Null()
    env = null.env


    allowed_methods = ('sart', 'sirt', 'asdpocs', 'sbtv')

    opts = kwargs['options']
    method_name = opts.get('method', 'sart')
    method_name = method_name if method_name in allowed_methods else 'sart'

    projector = env.pm.get_task_from_package('ir', 'parallel-projector')
    projector.set_properties(model='joseph', is_forward=False, axis_position=center, step=theta)

    method = env.pm.get_task_from_package('ir', method_name)
    method.set_properties(projector=projector, num_iterations=opts.get('num_iterations', 5))

    if method_name in ('sart', 'sirt'):
        method.set_properties(relaxation_factor=opts.get('relaxation_factor', 0.25))

    if method_name == 'asdpocs':
        minimizer = env.pm.get_task_from_package('ir', 'sirt')
        method.set_properties(df_minimizer=minimizer)

    if method_name == 'sbtv':
        method.set_properties(mu=opts.get('mu', 0.5))

    method = Task(env, 'ir', {}, method)

    for i, slce in enumerate(method(p.sinograms)):
        volume[i,:,:] = slce[:,:]


def ir(*args, **kwargs):
    """
    Reconstruct object using UFO's direct iterative reconstruction methods

    Extra options
    -------------
    method : str
        Any of `sart`, `sirt`, `asdpocs` and `sbtv`.

    num_iterations : int, optional
        Number of iterations before stopping reconstruction.

    relaxation_factor : float
        Relaxation factor for SART and SIRT methods.

    mu : float
        mu for Split Bregman Total Variation algorithm.

    Note
    ----
    You have to set ``ncore`` to 1 because UFO provides multi-threading on its
    own.
    """
    result = [_ir]
    result.extend(args)
    result.append(kwargs)
    return result



import os
import tempfile
import shutil
import contextlib


def disable(func):
    def empty():
        pass

    return empty


class TestDir(object):
    def __init__(self, path):
        self.root = path

    def path(self, name):
        return os.path.join(self.root, name)

    def __str__(self):
        return self.root


@contextlib.contextmanager
def tempdir():
    root = tempfile.mkdtemp()
    yield TestDir(root)
    shutil.rmtree(root)
